/* Copyright (C) 2010 Lucas Baudin
*
* This file is part of Ella Project.
*
* Ella Project is free software : you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* Ella Project is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY ; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Ella Project. If not, see <http ://www.gnu.org/licenses/>.
*/

#include "ella-user_config.h"

static GKeyFile* keyfile;
static gchar* config_file;

void ella_load_config(void)
{
    /* Creating a new keyfile object, to load the user config in it. */
    keyfile = g_key_file_new();

    /*
     * Build the path of the config file, usually, it is something like
     * ~/.config/ella/user.cfg
     * */
    config_file = g_build_filename(ella_dirs_get_user_config_dir(),
                                   "user.cfg",
                                   NULL);

    /* Load the keys, it can be empty, that's why we will put some default value
     * then. */
    g_key_file_load_from_file(keyfile, config_file,
                              G_KEY_FILE_NONE,
                              NULL);

    /*
     * Add default values.
     * TODO: it is hardcoded, not sure it is a good thing.
     * */
    ella_add_key_double("PAGE_WIDTH", 550.0);
    ella_add_key_double("PAGE_HEIGHT", 440.0);
    ella_add_key_int("DONT_ASK_BEFORE_QUIT", 1);
    ella_add_key_int("PAGE_COLOR_RED", 0xFFFF);
    ella_add_key_int("PAGE_COLOR_GREEN", 0xFFFF);
    ella_add_key_int("PAGE_COLOR_BLUE", 0xFFFF);
    ella_add_key_int("FPS", 12);
    ella_add_key_int("FOOTAGE", 512);
    ella_add_key_int("MARGIN", 80);
    ella_add_key_int("W_FEN", 1000);
    ella_add_key_int("H_FEN", 800);
    ella_add_key_int("TIMELINE_ZOOM", 75);
    ella_add_key_int("AUTO_ARCHIVE", 0);
    ella_add_key_int("EXPORT_BACKGROUND", 1);
    ella_add_key("RECENT_FILE_0", "");
    ella_add_key("RECENT_FILE_1", "");
    ella_add_key("RECENT_FILE_2", "");
    ella_add_key("RECENT_FILE_3", "");
    ella_add_key("RECENT_FILE_4", "");
    ella_add_key("RECENT_FILE_5", "");
}

void ella_add_key(const gchar* field, const gchar* value)
{
    /* It is used for default value, so, we must'nt change the value of an
     * existing key! */
    if(!g_key_file_has_key (keyfile, "Config", field, NULL))
    {
        g_key_file_set_value(keyfile, "Config", field, value);
    }
}

void ella_add_key_double(const gchar* field, double value)
{
    /* It is used for default value, so, we must'nt change the value of an
     * existing key! */
    if(!g_key_file_has_key (keyfile, "Config", field, NULL))
    {
        g_key_file_set_double(keyfile, "Config", field, value);
    }
}

void ella_add_key_int(const gchar* field, int value)
{
    /* It is used for default value, so, we must'nt change the value of an
     * existing key! */
    if(!g_key_file_has_key (keyfile, "Config", field, NULL))
    {
        g_key_file_set_integer(keyfile, "Config", field, value);
    }
}

gchar* ella_config_get_value(const gchar* string)
{
    return g_key_file_get_value(keyfile, "Config", string, NULL);
}

double ella_config_get_value_double(const gchar* string)
{
    return g_key_file_get_double(keyfile, "Config", string, NULL);
}

gint ella_config_get_value_int(const gchar* string)
{
    return g_key_file_get_double(keyfile, "Config", string, NULL);
}

void ella_config_set_value(const gchar* field, gchar* value)
{
    g_key_file_set_value(keyfile, "Config", field, value);
}

void ella_config_set_value_int(const gchar* field, int value)
{
    g_key_file_set_integer(keyfile, "Config", field, value);
}

void ella_config_set_value_double(const gchar* field, gdouble value)
{
    g_key_file_set_double(keyfile, "Config", field, value);
}

void ella_save_config(void)
{
    /* Save the config... */
    FILE* file_to_write;
    gchar* data;

    /* config_dir has been computed at the beginning, in ella_load_config. */
    file_to_write = fopen (config_file, "w");
    if (file_to_write)
    {
        data = g_key_file_to_data (keyfile, NULL, NULL);
        fputs (data, file_to_write);
        fclose (file_to_write);
        g_free (data);
    }
    else
    {
        g_warning ("Could not write configuration file: %s", config_file);
    }
}

GdkColor* ella_config_get_gtk_color(GtkWidget* widget)
{

    static GdkColor* theme_color;
    if(theme_color == NULL)
    {
        theme_color = g_malloc(sizeof *theme_color);

        /* Look up the default text color in the theme, use a default 
        if it's not defined */
        GtkStyle *style = gtk_rc_get_style(widget);
        if(!gtk_style_lookup_color(style, "selected_bg_color", theme_color))
            gdk_color_parse("black", theme_color);
    }
    return theme_color;
}

gdouble ella_config_get_gtk_color_r(GtkWidget* widget)
{
    return ((double)ella_config_get_gtk_color(widget)->red)/ 0xFFFF;
}

gdouble ella_config_get_gtk_color_g(GtkWidget* widget)
{
    return ((double)ella_config_get_gtk_color(widget)->green)/ 0xFFFF;
}

gdouble ella_config_get_gtk_color_b(GtkWidget* widget)
{
    return ((double)ella_config_get_gtk_color(widget)->blue)/ 0xFFFF;
}

GdkColor* ella_config_get_gtk_color_fg(GtkWidget* widget)
{

    static GdkColor* theme_color;
    if(theme_color == NULL)
    {
        theme_color = g_malloc(sizeof *theme_color);

        /* Look up the default text color in the theme, use a default 
        if it's not defined */
        GtkStyle *style = gtk_rc_get_style(widget);
        if(!gtk_style_lookup_color(style, "fg_color", theme_color))
            gdk_color_parse("black", theme_color);
    }
    return theme_color;
}

gdouble ella_config_get_gtk_color_fg_r(GtkWidget* widget)
{
    return ((double)ella_config_get_gtk_color_fg(widget)->red)/ 0xFFFF;
}

gdouble ella_config_get_gtk_color_fg_g(GtkWidget* widget)
{
    return ((double)ella_config_get_gtk_color_fg(widget)->green)/ 0xFFFF;
}

gdouble ella_config_get_gtk_color_fg_b(GtkWidget* widget)
{
    return ((double)ella_config_get_gtk_color_fg(widget)->blue)/ 0xFFFF;
}

GdkColor* ella_config_get_gtk_color_bg(GtkWidget* widget)
{

    static GdkColor* theme_color;
    if(theme_color == NULL)
    {
        theme_color = g_malloc(sizeof *theme_color);

        /* Look up the default text color in the theme, use a default 
        if it's not defined */
        GtkStyle *style = gtk_rc_get_style(widget);
        if(!gtk_style_lookup_color(style, "bg_color", theme_color))
            gdk_color_parse("black", theme_color);
    }
    return theme_color;
}

gdouble ella_config_get_gtk_color_bg_r(GtkWidget* widget)
{
    return ((double)ella_config_get_gtk_color_bg(widget)->red)/ 0xFFFF;
}

gdouble ella_config_get_gtk_color_bg_g(GtkWidget* widget)
{
    return ((double)ella_config_get_gtk_color_bg(widget)->green)/ 0xFFFF;
}

gdouble ella_config_get_gtk_color_bg_b(GtkWidget* widget)
{
    return ((double)ella_config_get_gtk_color_bg(widget)->blue)/ 0xFFFF;
}
