/* Copyright (C) 2011 Lucas Baudin
 *
 * This file is part of Ella Project.
 *
 * Ella Project is free software : you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Ella Project is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY ; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ella Project. If not, see <http ://www.gnu.org/licenses/>.
 */

public class Ella.Undo.Manager : Object
{
    Gee.ArrayList<Event> back_stack;
    Gee.ArrayList<Event> forward_stack;
    Event _current;
    public Event current_event {get { return _current; } }
    public Manager()
    {
        back_stack = new Gee.ArrayList<Event>();
        forward_stack = new Gee.ArrayList<Event>();
    }

    public void register(Event event)
    {
        forward_stack.clear();
        back_stack.add(event);
        _current = null;
    }

    public bool back()
    {
        if(back_stack.size > 0)
        {
            if(_current != null)
            {
                forward_stack.add(_current);
            }
            _current = back_stack[back_stack.size - 1];
            back_stack.remove(_current);
            _current.undo();
            return true;
        }
        return false;
    }

    public bool forward()
    {
        if(_current != null)
        {
            _current.redo();
            back_stack.add(_current);
            if(forward_stack.size > 0)
            {
                _current = forward_stack[forward_stack.size - 1];
                forward_stack.remove(_current);
            }
            else
                _current = null;
            return true;
        }
        return false;
    }
}

public abstract class Ella.Undo.Event : Object
{
    public abstract void undo ();
    public abstract void redo();
}

public abstract class Ella.Undo.EventItem<G> : Ella.Undo.Event
{
    Gee.ArrayList<Item> items;
    Gee.ArrayList<G> move;
    Gee.ArrayList<G> move_to;
    public void init_event(Gee.ArrayList<Item> items)
    {
        this.items = new Gee.ArrayList<Item>();
        this.items.add_all(items);
        move = new Gee.ArrayList<G>();
        move_to = new Gee.ArrayList<G>();
        foreach(var item in items)
        {
            fill(item, move);
        }
        update(items);
    }

    public void update(Gee.ArrayList<Item> items)
    {
        if(items.size != items.size)
        {
            print("Error in the undo manager");
        }
        move_to.clear();
        foreach(var item in items)
        {
            fill(item, move_to);
        }
    }

    public override void redo()
    {
        for(int i = 0; i < items.size; i++)
        {
            apply(items[i], move_to[i]);
        }
    }

    public override void undo()
    {
        for(int i = 0; i < items.size; i++)
        {
            apply(items[i], move[i]);
        }
    }

    public abstract void apply(Item item, G values);
    public abstract void fill(Item item, Gee.ArrayList<G> values);
}

public class FourDoubles
{
    public double value1;
    public double value2;
    public double value3;
    public double value4;
    public FourDoubles(double value1_, double value2_, double value3_, double value4_)
    {
        value1 = value1_;
        value2 = value2_;
        value3 = value3_;
        value4 = value4_;
    }
}

public class TwoDoubles
{
    public double value1;
    public double value2;
    public TwoDoubles(double value1_, double value2_)
    {
        value1 = value1_;
        value2 = value2_;
    }
}

public class Ella.Undo.EventMoveItem : Ella.Undo.EventItem<TwoDoubles>
{
    public EventMoveItem(Gee.ArrayList<Item> items)
    {
        init_event(items);
    }
    public override void apply(Item item, TwoDoubles values)
    {
        item.x = values.value1;
        item.y = values.value2;
    }

    public override void fill(Item item, Gee.ArrayList<TwoDoubles> value_list)
    {
        value_list.add(new TwoDoubles(item.x, item.y));
    }
}

public class Ella.Undo.EventResizeItem : Ella.Undo.EventItem<FourDoubles>
{
    public EventResizeItem(Gee.ArrayList<Item> items)
    {
        init_event(items);
    }
    public override void apply(Item item, FourDoubles values)
    {
        item.width = values.value1;
        item.height = values.value2;
        item.x = values.value3;
        item.y = values.value4;
    }

    public override void fill(Item item, Gee.ArrayList<FourDoubles> value_list)
    {
        value_list.add(new FourDoubles(item.width, item.height, item.x, item.y));
    }
}

public class Ella.Undo.EventRotateItem : Ella.Undo.EventItem<double?>
{
    public EventRotateItem(Gee.ArrayList<Item> items)
    {
        init_event(items);
    }

    public override void apply(Item item, double? value_)
    {
        item.rotation = value_;
    }

    public override void fill(Item item, Gee.ArrayList<double?> value_list)
    {
        value_list.add(item.rotation);
    }
}
