/* Copyright (C) 2011 Lucas Baudin
 *
 * This file is part of Ella Project.
 *
 * Ella Project is free software : you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Ella Project is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY ; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ella Project. If not, see <http ://www.gnu.org/licenses/>.
 */



using Gtk;
using Cairo;
using Gee;
using Ella;
const int BYTE_BITS = 8;

const int WORD_BITS = 16;
const int DWORD_BITS = 32;

/* Drag and drop */
private enum Target {
	STRING,
	URL,
	TYPE_PIXMAP
}

/* datatype (string), restrictions on DnD (Gtk.TargetFlags), datatype (int) */
private const TargetEntry[] target_list = {
      { "STRING",         0, Target.STRING },
      {"text/plain",      0, Target.STRING},
      {"text/uri-list",   0, Target.URL },
	  {"application/x-rootwindow-drop", 0, Target.TYPE_PIXMAP}
};


namespace Ella
{
    /**
     * The differents tool that the canvas can use.
     * They can be used in the canvas with the #ella_canvas_get_tool and #ella_canvas_set_tool functions.
     */
    public enum Tool
    {
        RECTANGLE,
        SELECT,
        CIRCLE,
        LINE,
        PATH,
        TEXT,
        IMAGE
    }

    private enum State
    {
        WAIT,
        SELECTING,
        MOVING,
        DRAWING,
        ROTATING,
        SCALING
    }
}

public class Ella.Canvas : DrawingArea
{
    /* ---- Private ---- */

    /* Our undo manager, it must be the same in all Ella code */
    private Undo.Manager undo_manager;
    private Undo.EventMoveItem event_move;
    private Undo.EventResizeItem event_resize;
    private Undo.EventRotateItem event_rotate;

    /* All the layers will be store in this list. */
    private ArrayList<Ella.Layer> layers_list;

    private Ella.State editor_state;

    /* Variable to save the mouse position */
    private double x_last_pressed = 0;
    private double y_last_pressed = 0;
    private double x_last_moved = 0;
    private double y_last_moved = 0;

    /* To recall if it was a single click or not, e.g. to switch from scaling to
     * rotating. */
    private bool m_change_tool = true;

    /* The height and the width of the little rect which are displayed when an
     * object is selected. */
    private double frame = 7;

    /* Frame position and number */
    private int frame_number = 0;
    private int max_keyframe = 0;

    private int layer_selected = -1;

    /* The color of the selection rectangle, colors are changed to fot with the
     * GTK+ theme. */
    private double select_red = 0;
    private double select_green = 0;
    private double select_blue = 0;

    /* The margins at the top and at the left, they are increaed when an object
     * is moved at the top, bottom, left, or right. */
    private double x_root;
    private double y_root;

    /* A flag which is set/unset when we press/release the control key. */
    private bool keep_ratio = false;

    /* A tool can have several subtoo, e.g. scaling/rotating. */
    private int tool_level;

    /* Which little rect is selected when we resize an object. */
    private int selected_frame = 0;


    /* All selected items will be store in this list, it is cleared and filled
     * at each selection. */
    private ArrayList<Ella.Item> item_selected;

    private bool is_editable = true;

    /* ---- Half private ----
     * 
     * They can be get and set in vala using the direct propertiy access e.g.
     * canvas.tool, but some actions can be done after they are set.
     *
     * In C, they can be set or get using a specific function, e.g.
     * ella_canvas_set_tool()
     *
     * Some times, we could just use public properties, but we wouldn't have
     * the nice C functions. (And use g_object_set is dirty, and it could be
     * broken if we need to extend a property.)
     *
     */

    private Ella.Tool _tool;
    public Ella.Tool tool
    {
        set { _tool = value; queue_draw(); }
        get { return _tool; }
    }

    private double _page_width = 800;
    public double page_width
    {
        get { return _page_width; }
        set { _page_width = value; compute_size_request(); }
    }

    private double _page_height = 550;
    public double page_height
    {
        get { return _page_height; }
        set { _page_height = value; compute_size_request(); }
    }

    private double _scale = 1.0;
    /**
     * Zoom level
     *
     * This is the scale/zoom level of the canvas, just change it to make a zoom
     * out/zoom in.
     */
    public double scale
    {
        get { return _scale; }
        set
        {
            _scale = value;
            compute_size_request();
            queue_draw();
        }
    }

    private double _page_color_red = 1.0;
    public double page_color_red
    {
        get { return _page_color_red; }
        set { _page_color_red = value; queue_draw(); }
    }

    private double _page_color_green = 1.0;
    public double page_color_green
    {
        get { return _page_color_green; }
        set { _page_color_green = value; queue_draw(); }
    }

    private double _page_color_blue = 1.0;
    public double page_color_blue
    {
        get { return _page_color_blue; }
        set { _page_color_blue = value; queue_draw(); }
    }
    

    /* ---- Public ---- */
    public signal void selection_changed();
    public signal void stop_play();
    public signal void frame_changed(int frame);
    public signal void export_png_finished();
    public signal void export_gif_finished();
    public signal void export_avi_finished();


    /**
     * New EllaCanvas widget.
     * 
     * Create a new EllaCanvas widget. It is derivated from a GtkDrawingArea.
     *
     */
    public Canvas()
    {
        Gdk.threads_init();
        add_events(Gdk.EventMask.BUTTON_PRESS_MASK
                  | Gdk.EventMask.BUTTON_RELEASE_MASK
                  | Gdk.EventMask.POINTER_MOTION_MASK
                  | Gdk.EventMask.KEY_PRESS_MASK
                  | Gdk.EventMask.KEY_RELEASE_MASK);

        set_can_focus(true);

        this.layers_list = new ArrayList<Ella.Layer>();
        //layers_list.add(new Layer());
        this.item_selected = new ArrayList<Ella.Item>();
        compute_size_request();
        
        Gtk.drag_dest_set(this,
                          Gtk.DestDefaults.ALL,
                target_list,
                Gdk.DragAction.COPY | Gdk.DragAction.MOVE | Gdk.DragAction.LINK
        );
        select_red = Ella.Config.get_gtk_color_r(this);
        select_green = Ella.Config.get_gtk_color_g(this);
        select_blue = Ella.Config.get_gtk_color_b(this);
    }

    public override bool focus(Gtk.DirectionType dir)
    {
        return false;
    }

    public void set_undo_manager(Undo.Manager manager)
    {
        this.undo_manager = manager;
    }

    public override bool key_press_event(Gdk.EventKey event)
    {
        double displace = 1;
        double displace_r = 10;
        switch(event.keyval)
        {
        case 65505:
            keep_ratio = true;
            break;
        case 65361:
            if(keep_ratio)
                move_selected_items(-displace_r, 0);
            else
                move_selected_items(-displace, 0);
            break;
        case 65362:
            if(keep_ratio)
                move_selected_items(0, -displace_r);
            else
                move_selected_items(0, -displace);
            break;
        case 65363:
            if(keep_ratio)
                move_selected_items(displace_r, 0);
            else
                move_selected_items(displace, 0);
            break;
        case 65364:
            if(keep_ratio)
                move_selected_items(0, displace_r);
            else
                move_selected_items(0, displace);
            break;
        default:
            print("Key pressed: " + event.keyval.to_string() + "\n");
            break;
        }
        return true;
    }

    public override bool key_release_event(Gdk.EventKey event)
    {
        switch(event.keyval)
        {
        case 65505:
            keep_ratio = false;
            break;
        case 65535:
            remove_selected_item();
            break;
        default:
            print("Key released: " + event.keyval.to_string() + "\n");
            break;
        }
        return true;
    }

    /**
     * Remove the item which are selected.
     */
    public void remove_selected_item()
    {
        switch(tool)
        {
        case Ella.Tool.SELECT:
            for(int i = 0; i < item_selected.size; i++)
            {
                if(!layers_list[layer_selected].remove_item(item_selected[i]))
                {
                    printerr("Couldn't remove an item, this should *not* happen.\n");
                }
            }
            item_selected.clear();
            break;
        case Ella.Tool.PATH:
            if(item_selected.size > 0 & item_selected[0] is Ella.Path)
            {
                ((Ella.Path)item_selected[0]).remove_point();
            }
            break;
        }
        queue_draw();
    }

    /**
     * Select all items
     */
    public void select_all_items()
    {
        item_selected.clear();
        layers_list[layer_selected].select_all(item_selected);
        queue_draw();
    }

    /**
     * Select all items
     */
    public void unselect_all_items()
    {
        item_selected.clear();
        queue_draw();
    }

    /**
     * Press expose event.
     *
     * This function is called by Gtk, don't use it!
     */
    public override bool button_press_event(Gdk.EventButton event)
    {
        if(!is_editable)
            return true;
        grab_focus();
        event.x = event.x / scale;
        event.y = event.y / scale;
        m_change_tool = false;
        switch(tool)
        {
        case Tool.RECTANGLE:
            var item = new Ella.Rectangle();
            item.height = 1;
            item.width = 1;
            item.x = event.x + x_root;
            item.y = event.y + y_root;
            layers_list[layer_selected].add_item(item);
            editor_state = Ella.State.SCALING;
            item_selected.clear();
            item_selected.add(item);
            selected_frame = 4;
            queue_draw();
            x_last_moved = event.x;
            y_last_moved = event.y;
            x_last_pressed = event.x;
            y_last_pressed = event.y;
            item.red = 1;
            selection_changed();
            break;
        case Tool.CIRCLE:
            var item = new Ella.Circle();
            item.height = 1;
            item.width = 1;
            item.x = event.x + x_root;
            item.y = event.y + y_root;
            layers_list[layer_selected].add_item(item);
            editor_state = Ella.State.SCALING;
            item_selected.clear();
            item_selected.add(item);
            selected_frame = 4;
            queue_draw();
            x_last_moved = event.x;
            y_last_moved = event.y;
            x_last_pressed = event.x;
            y_last_pressed = event.y;
            item.red = 1;
            selection_changed();
            break;
        case Tool.SELECT:
            /* Save the position */

            x_last_pressed = event.x;
            y_last_pressed = event.y;
            x_last_moved = event.x;
            y_last_moved = event.y;
            m_change_tool = true;
            //item_selected.clear();
            if(!is_in_small_frames(event.x, event.y) && !is_in_selection(event.x, event.y))
            {
                m_change_tool = false;
                select_here(event.x, event.y);
            }

            if(item_selected.size > 0)
            {
                if(is_in_small_frames(event.x, event.y))
                {
                    if(tool_level == 0)
                    {
                        editor_state = Ella.State.SCALING;
                        event_resize = new Ella.Undo.EventResizeItem(item_selected);
                    }
                    else
                    {
                        editor_state = Ella.State.ROTATING;
                        event_rotate = new Ella.Undo.EventRotateItem(item_selected);
                    }
                }
                else
                {
                    event_move = new Ella.Undo.EventMoveItem(item_selected);
                    editor_state = Ella.State.MOVING;
                }
            }
            else
            {
                editor_state = Ella.State.SELECTING;
            }
            break;
        case Tool.PATH:
            Ella.Path item;
            if(item_selected.size == 0 || (item_selected.size > 0 & ! (item_selected[0] is Ella.Path)))
            {
                item = new Ella.Path();
                layers_list[layer_selected].add_item(item);
                item_selected.clear();
                item.red = 1.0;
                item.height = 1;
                item.width = 1;
                item.red_stroke = 1.0;
                item.line_width = 1.0;
                item_selected.add(item);
                item.type_item = Ella.ItemType.PATH;
                item.x = event.x + x_root;
                item.y = event.y + y_root;
                item.add_point(0, 0);
                var point = item.add_point(0, 0);
                item.select_point(point);
            }
            else
            {
                item = (Ella.Path?)item_selected[0];
                item.select_point_at(event.x + x_root, event.y + y_root);
            }
            editor_state = Ella.State.MOVING;
            break;
        case Tool.TEXT:
            var item = new Ella.Text();
            item.x = event.x + x_root;
            item.y = event.y + y_root;
            item.type_item = Ella.ItemType.TEXT;
            layers_list[layer_selected].add_item(item);
            editor_state = Ella.State.WAIT;
            item_selected.clear();
            item_selected.add(item);
            x_last_moved = event.x;
            y_last_moved = event.y;
            x_last_pressed = event.x;
            y_last_pressed = event.y;
            item.red = 0;
            item.alpha = 1;
            item.green = 0;
            item.blue = 0;
            item.text = "<Now, type the text in the right panel>";
            item.height = 100;
            item.width = 100;
            item.height = 50;
            item.width = 50;
            /* Let's redraw it twice to avoid any visual glitchs. */
            queue_draw();
            queue_draw();
            selection_changed();
            break;
        }
        return false;
    }

    /**
     * Release expose event.
     *
     * This function is called by Gtk, don't use it!
     */
    public override bool button_release_event(Gdk.EventButton event)
    {
        event.x = event.x / scale;
        event.y = event.y / scale;
        /* TODO: this function isn't finshed at all! */
        /* Is it just a single click? */
        
        if(x_last_pressed - 5 < event.x && event.x < x_last_pressed + 5 && y_last_pressed - 5 < event.y && event.y < y_last_pressed + 5)
        {
            if(m_change_tool)
            {
                /* Yeah, it is a single click :) */
                if(tool_level < 1)
                {
                    tool_level++;
                }
                else
                    tool_level = 0;
            }
        }

        switch(editor_state)
        {
            case Ella.State.SELECTING:
            {
                if(x_last_pressed - 5 < event.x && event.x < x_last_pressed + 5 && y_last_pressed - 5 < event.y && event.y < y_last_pressed + 5)
                {
                    select_here(event.x, event.y);
                }
                else
                {
                    select_area(x_last_pressed, y_last_pressed, event.x, event.y);
                }
                editor_state = Ella.State.WAIT;
                break;
            }
            case Ella.State.MOVING:
            {
                undo_manager.register(event_move);
                event_move = null;
                editor_state = Ella.State.WAIT;
                break;
            }
            case Ella.State.ROTATING:
            {
                editor_state = Ella.State.WAIT;
                undo_manager.register(event_rotate);
                event_rotate = null;

                break;
            }
            case Ella.State.SCALING:
            {
                editor_state = Ella.State.WAIT;
                undo_manager.register(event_resize);
                event_resize = null;
                break;
            }
        }
        queue_draw();
        return false;
    }

    private void move_selected_items(double x, double y)
    {
        
        for(int i = 0; i < item_selected.size; i++)
        {
            item_selected[i].x += x;
            item_selected[i].y += y;
        }
        event_move.update(item_selected);
        compute_size_request();
        queue_draw();
    }

    /**
     * Motion expose event.
     *
     * This function is called by Gtk, don't use it!
     */
    public override bool motion_notify_event(Gdk.EventMotion event)
    {
        if(!is_editable)
            return true;
        event.x = event.x / scale;
        event.y = event.y / scale;
        switch(editor_state)
        {
            case Ella.State.MOVING:
            {
                switch(tool)
                {
                case Tool.PATH:
                    var item = (Ella.Path?)item_selected[0];
                    item.move_point(event.x + x_root - item.x, event.y + y_root - item.y);
                    compute_size_request();
                    queue_draw();
                    x_last_moved = event.x;
                    y_last_moved = event.y;
                    break;
                default:
                    move_selected_items(event.x - x_last_moved, event.y - y_last_moved);
                    x_last_moved = event.x;
                    y_last_moved = event.y;
                    break;
                }
                break;
            }
            case Ella.State.SCALING:
            {
                Rect area = get_selection_area();
                double  x = (event.x - x_last_moved);
                double y = (event.y - y_last_moved);
                for(int i = 0; i < item_selected.size; i++)
                {
                    double old_height = item_selected[i].height;
                    double old_width = item_selected[i].width;
                    switch(selected_frame)
                    {
                    case 0:
                        item_selected[i].x += x*(area.width + area.x - item_selected[i].x)/area.width;
                        item_selected[i].y += y*(area.height + area.y - item_selected[i].y)/area.height;
                        item_selected[i].width -= x*(item_selected[i].width / area.width);
                        item_selected[i].height -= y*(item_selected[i].height)/(area.height);
                        if(item_selected[i].width <= 0 && item_selected[i].height <= 0)
                        {
                            item_selected[i].width = 1;
                            item_selected[i].height = 1;
                            selected_frame = 4;
                        }
                        else if(item_selected[i].width <= 0)
                        {
                            item_selected[i].width = 1;
                            selected_frame = 2;
                        }
                        else if(item_selected[i].height <= 0)
                        {
                            item_selected[i].height = 1;
                            selected_frame = 6;
                        }
                        break;
                    case 2:
                        item_selected[i].x -= x*(area.x - item_selected[i].x)/area.width;
                        item_selected[i].y += y*(area.height + area.y - item_selected[i].y)/area.height;
                        item_selected[i].height -= y*(item_selected[i].height)/(area.height);
                        item_selected[i].width += x*(item_selected[i].width / area.width);
                        if(item_selected[i].width <= 0 && item_selected[i].height <= 0)
                        {
                            item_selected[i].width = 1;
                            item_selected[i].height = 1;
                            selected_frame = 6;
                        }
                        else if(item_selected[i].width <= 0)
                        {
                            item_selected[i].width = 1;
                            selected_frame = 0;
                        }
                        else if(item_selected[i].height <= 0)
                        {
                            item_selected[i].height = 1;
                            selected_frame = 4;
                        }
                        break;
                    case 4:
                        item_selected[i].x -= x*(area.x - item_selected[i].x)/area.width;
                        item_selected[i].y -= y*(area.y - item_selected[i].y)/area.height;
                        item_selected[i].width += x*(item_selected[i].width / area.width);
                        item_selected[i].height += y*(item_selected[i].height / area.height);
                        if(item_selected[i].width <= 0 && item_selected[i].height <= 0)
                        {
                            item_selected[i].width = 1;
                            item_selected[i].height = 1;
                            selected_frame = 0;
                        }
                        else if(item_selected[i].width <= 0)
                        {
                            item_selected[i].width = 1;
                            selected_frame = 6;
                        }
                        else if(item_selected[i].height <= 0)
                        {
                            item_selected[i].height = 1;
                            selected_frame = 2;
                        }
                        break;
                    case 6:
                        item_selected[i].x += x*(area.width + area.x - item_selected[i].x)/area.width;
                        item_selected[i].y -= y*(area.y - item_selected[i].y)/area.height;
                        item_selected[i].width -= x*(item_selected[i].width / area.width);
                        item_selected[i].height += y*(item_selected[i].height / area.height);
                        if(item_selected[i].width <= 0 && item_selected[i].height <= 0)
                        {
                            item_selected[i].width = 0;
                            item_selected[i].height = 0;
                            selected_frame = 2;
                        }
                        else if(item_selected[i].width <= 0)
                        {
                            item_selected[i].width = 0;
                            selected_frame = 4;
                        }
                        else if(item_selected[i].height <= 0)
                        {
                            item_selected[i].height = 0;
                            selected_frame = 0;
                        }
                        break;
                    }
                    
                    /* To keep the ratio */
                    if(keep_ratio)
                    {
                        /* Change the height proportionaly to the width change */
                        item_selected[i].height = old_height + old_height * (item_selected[i].width - old_width) / old_width;
                    }
                    if(item_selected[i].height == 0)
                    {
                        item_selected[i].height = old_height;
                    }
                    if(item_selected[i].width == 0)
                    {
                        item_selected[i].width = old_width;
                    }
                    if(event_resize != null) /* yeah, it can be null if it is a new object */
                    {
                        event_resize.update(item_selected);
                    }
                }
                queue_draw();
                x_last_moved = event.x;
                y_last_moved = event.y;
                break;
            }
            case Ella.State.ROTATING:
            {
                foreach(Item item in item_selected)
                {
                    double angle = ComputeAngle(item.get_center()[0], item.get_center()[1], x_last_moved + x_root, y_last_moved + y_root, event.x + x_root, event.y + y_root);
                    item.rotation += angle;
                }
                event_rotate.update(item_selected);
                x_last_moved = event.x;
                y_last_moved = event.y;
                queue_draw();
                break;
            }
            case Ella.State.SELECTING:
            {
                x_last_moved = event.x;
                y_last_moved = event.y;
                queue_draw();
                break;
            }
        }
        return true;
    }

    public void up_selected_item()
    {
        if(item_selected.size > 0)
        {
            layers_list[layer_selected].item_up_instance(item_selected[0]);
            queue_draw();
        }
    }

    public void down_selected_item()
    {
        if(item_selected.size > 0)
        {
            layers_list[layer_selected].item_down_instance(item_selected[0]);
            queue_draw();
        }
    }


    private void draw_canvas(Context cr)
    {
        cr.scale(_scale, _scale);
        draw_page(cr);
        draw_all_items(cr);
        draw_selection(cr);
    }
    string path_png = "/tmp/ella_";
    string path_gif = "/tmp/ella_";
    string path_avi = "/tmp/ella_";

    /**
     * Export to PNG, in /tmp/ella_xxx.png (replace xxx by the frame number).
     */
    public void export(string path_png)
    {
        double old_x_root = x_root;
        double old_y_root = y_root;
        bump_maxkeyframe();
        x_root = 0;
        y_root = 0;
        for(int i = 0; i < max_keyframe; i++)
        {
            ImageSurface image = new ImageSurface(Format.ARGB32, (int)_page_width, (int)_page_height);
            change_keyframe(i);
            Context context = new Context(image);

            context.scale(_scale, _scale);
            draw_page(context);
            draw_all_items(context);

            if(i < 10)
                image.write_to_png(path_png + "00" + i.to_string() + ".png");
            else if(i < 100)
                image.write_to_png(path_png + "0" + i.to_string() + ".png");
            else
                image.write_to_png(path_png + i.to_string() + ".png");
        }
        x_root = old_x_root;
        y_root = old_y_root;
    }
    int movie_frequence;
    public void export_to_gif()
    {
        export("/tmp/ella_gif_");
        string command = "convert -delay %d -loop 0 /tmp/ella_gif_*.png %s".printf(100/movie_frequence, path_gif);
        //print(command + "\n");
        Process.spawn_command_line_sync(command);
        export_gif_finished();
    }
    public void export_to_avi()
    {
        export("/tmp/ella_avi_");
        string command = "mencoder mf:///tmp/ella_avi_*.png -mf w=%d:h=%d:fps=%d:type=png -ovc lavc -lavcopts vcodec=mpeg4 -oac copy -o %s".printf((int)page_width, (int)page_height, movie_frequence, path_avi);
        print(command + "\n");
        Process.spawn_command_line_sync(command);
        export_avi_finished();
    }
 
    public void export_to_png()
    {
        print("Export png to: %s\n", path_png);
        export(path_png);
        export_png_finished();
    }
  
    public void export_to_png_async(string path)
    {
        path_png = path;
        unowned Thread thread = Thread.create<void>(export_to_png, false);
    }

    public void export_to_gif_async(string path, int movie_frequence)
    {
        path_gif = path;
        this.movie_frequence = movie_frequence;
        unowned Thread thread = Thread.create<void>(export_to_gif, false);
    }

    public void export_to_avi_async(string path, int movie_frequence)
    {
        path_avi = path;
        this.movie_frequence = movie_frequence;
        unowned Thread thread = Thread.create<void>(export_to_avi, false);
    }

    private void draw_selection(Context cr)
    {

        /* Selection */
        /* FIXME: this should be moved in an other file, and be more coded
         * to have nice effects. */
        if(item_selected.size > 0)
        {
            double dashed1[2];
            dashed1[0] = 4.0;
            dashed1[1] = 4.0;

            cr.set_dash(dashed1, 0);
            cr.set_line_width(2);
            Rect area = get_selection_area();
            area.x -= x_root;
            area.y -= y_root;
            cr.rectangle(area.x,
                         area.y,
                         area.width,
                         area.height);
            if(tool_level == 0)
                cr.set_source_rgb(1, 0, 0);
            else
                cr.set_source_rgb(0, 1, 0);
            cr.stroke();
            /* top left */
            if(is_editable)
                cr.set_source_rgb(0, 0, 0);
            else
                cr.set_source_rgb(0.7, 0.7, 0.7);
            cr.rectangle(area.x - frame/2, area.y - frame/2, frame, frame);
            cr.rectangle(area.x - frame/2, area.y + area.height - frame/2, frame, frame);
            cr.rectangle(area.x + area.width - frame/2, area.y - frame/2, frame, frame);
            cr.rectangle(area.x + area.width - frame/2, area.y + area.height - frame/2, frame, frame);
            cr.rectangle(area.x + area.width/2- frame/2, area.y - frame/2, frame, frame);
            cr.rectangle(area.x + area.width/2 - frame/2, area.y + area.height - frame/2, frame, frame);
            cr.rectangle(area.x - frame/2, area.y + area.height/2 - frame/2, frame, frame);
            cr.rectangle(area.x + area.width - frame/2, area.y + area.height/2 - frame/2, frame, frame);
            cr.fill();
        }
        if(editor_state == Ella.State.SELECTING)
        {
            double dashed1[2];
            dashed1[0] = 4.0;
            dashed1[1] = 4.0;

            cr.set_dash(dashed1, 0);
            cr.set_line_width(2);
            cr.rectangle(x_last_pressed,
                         y_last_pressed,
                         x_last_moved - x_last_pressed,
                         y_last_moved - y_last_pressed);
            cr.set_source_rgb(select_red, select_green, select_blue);
            cr.stroke();
        }
        switch(tool)
        {
        case Tool.PATH:
            foreach(Item item in item_selected)
            {
                item.draw_editing(cr, x_root, y_root);
            }
            break;
        }
    }

    private void draw_all_items(Context cr)
    {
        foreach(Layer layer in layers_list)
        {
            layer.draw_all_items(cr, x_root, y_root);
        }
    }

    private void draw_page(Context cr)
    {
        cr.rectangle(- x_root + 3, - y_root + 3, page_width, page_height);
        cr.set_source_rgb(0.2, 0.2, 0.2);
        cr.fill();
        cr.rectangle(- x_root, - y_root, page_width, page_height);
        cr.set_source_rgb(page_color_red, page_color_green, page_color_blue);
        cr.fill_preserve();
        cr.set_source_rgb(0.2, 0.2, 0.2);
        cr.set_line_width(1);
        cr.stroke();
    }

    private void select_area(double x, double y, double x2, double y2)
    {
        Ella.Utils.sort_coords(ref x, ref y, ref x2, ref y2);
        item_selected.clear();
        x += x_root;
        x2 += x_root;
        y += y_root;
        y2 += y_root;
        layers_list[layer_selected].select_area(item_selected, x, y, x2, y2);
        selection_changed();
    }

    private bool is_in_selection(double x, double y)
    {
        x += x_root;
        y += y_root;
        for(int i = 0; i < item_selected.size; i++)
        {
            Rect area = item_selected[i].Area();
            if(x + frame > area.x && x - frame < area.x + area.width)
            {
                if(y + frame > area.y && y - frame < area.y + area.height)
                {
                    return true;
                }
            }
        }
        return false;
    }

    private bool is_in_small_frames(double x, double y)
    {
        x += x_root;
        y += y_root;
        Rect area = get_selection_area();
        Rect frame1 = Rect();
        frame1.x = area.x - frame/2;
        frame1.y = area.y - frame/2;
        frame1.width = frame;
        frame1.height = frame;
        if(RectIsIn(x, y, frame1))
        {
            selected_frame = 0;
            return true;
        }

        frame1.x = area.x + area.width - frame/2;
        frame1.y = area.y - frame/2;
        if(RectIsIn(x, y, frame1))
        {
            selected_frame = 2;
            return true;
        }

        frame1.x = area.x - frame/2;
        frame1.y = area.y + area.height - frame/2;
        if(RectIsIn(x, y, frame1))
        {
            selected_frame = 6;
            return true;
        }

        frame1.x = area.x + area.width - frame/2;
        frame1.y = area.y + area.height - frame/2;
        if(RectIsIn(x, y, frame1))
        {
            selected_frame = 4;
            return true;
        }

        return false;
    }

    private Rect get_selection_area()
    {
        Rect area = Rect();
        double x1 = 12800;
        double y1 = 12800;
        double y2 = -12800;
        double x2 = -12800;
        area.x = x1;
        area.y = 1280;
        area.width = 0;
        area.height = 0;
        for(int i = 0; i < item_selected.size; i++)
        {
            Rect area_item = item_selected[i].Area();
            if(area_item.x < x1)
            {
                x1 = area_item.x;
            }
            if(area_item.y < y1)
            {
                y1 = area_item.y;
            }
            if(x2 < area_item.x + area_item.width)
            {
                x2 = area_item.x + area_item.width;
            }
            if(y2 < area_item.y + area_item.height)
            {
                y2 = area_item.y + area_item.height;
            }
        }
        area.x = x1;
        area.y = y1;
        area.width = x2 - x1;
        area.height = y2 - y1;
        return area;
    }

    private bool select_here(double x, double y)
    {
        if(!is_in_selection(x, y))
        {
            item_selected.clear();
            x += x_root;
            y += y_root;
            layers_list[layer_selected].select_here(item_selected, x, y);
            if(item_selected.size > 0)
            {
                editor_state = Ella.State.WAIT;
            }
            selection_changed();
            return false;
        }
        return true;
    }

    public override bool draw(Cairo.Context cr)
    {
        cr.rectangle (0, 0,
                      width_request, height_request);
        cr.clip();
        cr.set_source_rgb(1, 1, 1);
        cr.paint();
        draw_canvas(cr);
        return true;
    }

    /**
     * Set the color selection.
     *
     * Set the color of all the items of the current selection.
     * @param r red channel, between 0.0 and 1.0
     * @param g red channel, between 0.0 and 1.0
     * @param b red channel, between 0.0 and 1.0
     * @param a alpha channel, between 0.0 and 1.0
     *
     */
    public void set_selection_color(double r, double g, double b, double a)
    {
        for(int i = 0; i < item_selected.size; i++)
        {
            item_selected[i].red = r;
            item_selected[i].green = g;
            item_selected[i].blue = b;
            item_selected[i].alpha = a;
        }
        queue_draw();
    }
    /**
     * Get the color if the first selected item.
     */
    public Ella.Color get_selection_color()
    {
        var color = Ella.Color();
        if(item_selected.size > 0)
        {
            color.red = item_selected[0].red;
            color.green = item_selected[0].green;
            color.blue = item_selected[0].blue;
            color.alpha = item_selected[0].alpha;
        }
        return color;
    }
    /**
     * Get the stroke color if the first selected item.
     */
    public Ella.Color get_selection_stroke_color()
    {
        var color = Ella.Color();
        if(item_selected.size > 0)
        {
            color.red = item_selected[0].red_stroke;
            color.green = item_selected[0].green_stroke;
            color.blue = item_selected[0].blue_stroke;
            color.alpha = item_selected[0].alpha_stroke;
        }
        return color;
    }

    /**
     * Get the color if the first selected item.
     */
    public double get_selection_line_width()
    {
        if(item_selected.size > 0)
        {
            return item_selected[0].line_width;
        }
        return 0;
    }

    /**
     * Set the color stroke selection.
     *
     * Set the stroke color of all the items of the current selection.
     * @param r red channel, between 0.0 and 1.0
     * @param g red channel, between 0.0 and 1.0
     * @param b red channel, between 0.0 and 1.0
     * @param a alpha channel, between 0.0 and 1.0
     */
    public void set_selection_stroke_color(double r, double g, double b, double a)
    {
        for(int i = 0; i < item_selected.size; i++)
        {
            item_selected[i].red_stroke = r;
            item_selected[i].green_stroke = g;
            item_selected[i].blue_stroke = b;
            item_selected[i].alpha_stroke = a;
        }
        queue_draw();
    }

    public void set_selection_line_width(double line_width)
    {
        for(int i = 0; i < item_selected.size; i++)
        {
            item_selected[i].line_width = line_width;
        }
        queue_draw();
    }

    public void set_selection_text(string text)
    {
        for(int i = 0; i < item_selected.size; i++)
        {
            item_selected[i].text = text;
        }
        queue_draw();
    }

    public void set_selection_font_name(string text)
    {
        foreach(var item in item_selected)
        {
            if(item is Ella.Text)
            {
                ((Ella.Text)item).font_name = text;
            }
        }
        queue_draw();
    }

    public string get_selection_text()
    {
        foreach(Item item in item_selected)
        {
            if(item is Text)
                return item.text;
        }
        return "";
    }

    public bool is_text_item_selected()
    {
        foreach(Item item in item_selected)
        {
            if(item is Text)
            {
                return true;
            }
        }
        return false;
    }
    private void compute_size_request()
    {
        if(layer_selected != -1)
        {
            double[] size_request = layers_list[layer_selected].compute_size_request(frame, page_width, page_height);
            width_request = (int)(size_request[1] * scale - size_request[0] * scale);
            y_root = (int)size_request[2];
            height_request = (int)(size_request[3] * scale - size_request[2] * scale);
            x_root = (int)size_request[0];
        }
    }

    /**
     * Change Keyframe.
     */
    public void change_keyframe(int time)
    {
        frame_number = time;
        foreach(Ella.Layer layer in layers_list)
        {
            layer.change_keyframe(time);
        }
        is_editable = layers_list[layer_selected].is_editable(time);
        queue_draw();
    }

    public int get_frame_number()
    {
        return frame_number;
    }

    /**
     * Add a keyframe to all objects.
     * @param time the position of the keyframe
     */
    public void add_keyframe(int time)
    {
        for(int i = 0; i < layers_list.size; i++)
        {
            layers_list[i].add_keyframe(time);
        }
    }
    
    public void bump_maxkeyframe()
    {
        for(int i = 0; i < layers_list.size; i++)
            max_keyframe = int.max(max_keyframe, layers_list[i].max_frames);
    }

    private bool is_playing = false;

    /**
     * Play the animation.
     *
     * Play the animation, with all the items.
     * @param start The frame where we must start the playing. If it is -1, we
     * start from the last position.
     * @param speed the speed of the animation, in miliseconds. e.g. if you have
     * 12 images per seconds, you must use 1000/12 for this parameter.
     */
    public void play(int start = -1, uint speed)
    {
        if(is_playing)
            return;
        is_playing = true;
        bump_maxkeyframe();
        /* So, we must start here at this frame. */
        if(start != -1)
            change_keyframe(start);
        else if(frame_number >= max_keyframe)
        {
            change_keyframe(0);
            frame_changed(frame_number);
        }
        Timeout.add(speed, () =>
        {
            /* Don't make an infinite loop: if the animation ended, just stop. */
            if(frame_number + 1 > max_keyframe || !is_playing)
            {
                is_playing = false;
                stop_play();
                return false;
            }
            else
            {
                change_keyframe(frame_number + 1);
                frame_changed(frame_number);
                return true;
            }
        });
    }
    /**
     * Stop the animation.
     */
    public void stop()
    {
        change_keyframe(0);
        frame_changed(frame_number);
        is_playing = false;
    }

    /**
     * Pause the animation.
     */
    public void pause()
    {
        is_playing = false;
    }
    
    
    public bool get_is_playing()
    {
        return is_playing;
    }

    public void add_layer(Layer layer)
    {
        layers_list.add(layer);
        compute_size_request();
        queue_draw();
    }

    public void clear_all_layers()
    {
        layer_selected = 0;
        layers_list.clear();
    }

    public void select_layer(Ella.Layer layer)
    {
        for(int i = 0; i < layers_list.size; i++)
        {
            if(layers_list[i] == layer)
            {
                layer_selected = i;
                compute_size_request();
                return;
            }
        }
        print("Couldn't find this layer, added it\n");
        layers_list.add(layer);
        layer_selected = layers_list.size - 1;
        item_selected.clear();
        queue_draw();
    }

    public override void drag_data_received(Gdk.DragContext context, int x, int y, Gtk.SelectionData data, uint target, uint time_) 
    {
        print("Dragdatareceived\n");
        var uris = data.get_text().split("\n");
        for(int i = 0; i < uris.length; i++)
        {
            var path = uris[i];
            path = path.replace("\r", "");
            path = path.replace("file://", "");
            if(path != "")
            {
                print("Trying to import file: " + path + "\n");
                import_png(path, x + x_root, y + y_root);
            }
        }
    }

    public void import_png(string path, double x=0, double y=0)
    {
        var item = new Image();
        item.height = 1;
        item.width = 1;
        item.x = x;
        item.y = y;
        item.set_image(path);
        layers_list[layer_selected].add_item(item);
        queue_draw();
    }

    public void layer_up(Ella.Layer layer)
    {
        int index = layers_list.index_of(layer);
       
        if(index <= 0)
            return;

        layers_list[index] = layers_list[index - 1];
        layers_list[index - 1] = layer;
        queue_draw();
    }

    public void layer_down(Ella.Layer layer)
    {
        int index = layers_list.index_of(layer);
       
        if(index >= layers_list.size - 1)
            return;

        layers_list[index] = layers_list[index + 1];
        layers_list[index + 1] = layer;
        queue_draw();
    }

    public void layer_remove(Ella.Layer layer)
    {
        if(layer == layers_list[layer_selected])
            layer_selected = 0;
        layers_list.remove(layer);
        queue_draw();
    }
}
