/* Copyright (C) 2010 Lucas Baudin
 *
 * This file is part of Ella Project.
 *
 * Ella Project is free software : you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Ella Project is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY ; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ella Project. If not, see <http ://www.gnu.org/licenses/>.
 */

#include "ella-color_selection_channel.h"

G_DEFINE_TYPE (EllaColorSelectionChannel, ella_color_selection_channel, GTK_TYPE_DRAWING_AREA);

#define ELLA_COLOR_SELECTION_CHANNEL_HEIGHT 25
#define ELLA_COLOR_SELECTION_CHANNEL_WIDTH 120

/* These functions handle events, define them in the header file is useless IMO,
 * since they _mustn't_ be called in an other file. */
static gboolean ella_color_selection_channel_button_press   (GtkWidget* widget, GdkEventButton* event);
static gboolean ella_color_selection_channel_button_motion  (GtkWidget* widget, GdkEventMotion* event);
static gboolean ella_color_selection_channel_button_release (GtkWidget* widget, GdkEventButton* event);

/* To paint the widget */
static gboolean ella_color_selection_channel_paint(GtkWidget* widget, cairo_t* cr);

enum
{
    VALUE_CHANGED,
    LAST_SIGNAL
};

static guint ella_color_selection_channel_signals[LAST_SIGNAL] = { 0 };

static void ella_color_selection_channel_class_init(EllaColorSelectionChannelClass* class)
{
    GtkWidgetClass* widget_class;
    widget_class = GTK_WIDGET_CLASS(class);

    widget_class->draw         = ella_color_selection_channel_paint;
    widget_class->button_press_event   = ella_color_selection_channel_button_press;
    widget_class->button_release_event = ella_color_selection_channel_button_release;
    widget_class->motion_notify_event  = ella_color_selection_channel_button_motion;


    /* Add the value-changed signal */
    ella_color_selection_channel_signals[VALUE_CHANGED] = g_signal_new(
        "value-changed",
        G_OBJECT_CLASS_TYPE (widget_class),
        G_SIGNAL_RUN_FIRST,
        G_STRUCT_OFFSET (EllaColorSelectionChannelClass, value_changed),
        NULL, NULL,
        g_cclosure_marshal_VOID__VOID,
        G_TYPE_NONE, 0);

}

static gboolean ella_color_selection_channel_button_release(GtkWidget* widget, GdkEventButton* event)
{
    /* We stop the drag mode */
    ELLA_COLOR_SELECTION_CHANNEL(widget)->is_pressed = FALSE;
    return TRUE;
}

#define ECSC_WIDTH gtk_widget_get_allocated_width(widget)

static gboolean ella_color_selection_channel_button_press(GtkWidget* widget, GdkEventButton* event)
{
    EllaColorSelectionChannel* color = ELLA_COLOR_SELECTION_CHANNEL(widget);
    /* We start the drag mode */
    color->is_pressed = TRUE;

    /* Convert the x mouse postion into a proper color value, 0 < x width, and
     * we need a color > 0 and < 1. */
    switch(color->channel)
    {
        case ELLA_COLOR_CHANNEL_RED:
        {
            color->red = event->x / ECSC_WIDTH;
            break;
        }
        case ELLA_COLOR_CHANNEL_BLUE:
        {
            color->blue = event->x / ECSC_WIDTH;
            break;
        }
        case ELLA_COLOR_CHANNEL_GREEN:
        {
            color->green = event->x / ECSC_WIDTH;
            break;
        }
    }

    /* Let's re-paint our widget :). */
    gtk_widget_queue_draw(widget);

    /* Emit the signal to say that the widget value changed. */
    g_signal_emit(widget,
                  ella_color_selection_channel_signals[VALUE_CHANGED],
                  0,
                  NULL); 

    return TRUE;
}

static gboolean ella_color_selection_channel_button_motion(GtkWidget* widget,
                                                           GdkEventMotion* event)
{
    if(ELLA_COLOR_SELECTION_CHANNEL(widget)->is_pressed)
    {
        gdouble value = event->x/ECSC_WIDTH;
        if(event->x < 0)
        {
            value = 0;
        }
        if(event->x > ECSC_WIDTH)
        {
            value = 1;
        }
        switch(ELLA_COLOR_SELECTION_CHANNEL(widget)->channel)
        {
            case ELLA_COLOR_CHANNEL_RED:
            {
                ELLA_COLOR_SELECTION_CHANNEL(widget)->red = value;
                break;
            }
            case ELLA_COLOR_CHANNEL_BLUE:
            {
                ELLA_COLOR_SELECTION_CHANNEL(widget)->blue = value;
                break;
            }
            case ELLA_COLOR_CHANNEL_GREEN:
            {
                ELLA_COLOR_SELECTION_CHANNEL(widget)->green = value;
                break;
            }
        }
        gtk_widget_queue_draw(widget);
        g_signal_emit(widget,
                      ella_color_selection_channel_signals[VALUE_CHANGED],
                      0,
                      NULL); 
    }
    return TRUE;
}

static gboolean ella_color_selection_channel_paint(GtkWidget* widget, cairo_t* cr)
{
    EllaColorSelectionChannel* color = ELLA_COLOR_SELECTION_CHANNEL(widget);

    GdkColor* theme_color; /* It is the default color of the user theme, to have
                            * a nice widget which fit with the user theme. */

    gdouble r, g, b; /* the values of where the gradient starts. */ 
    gdouble r_, g_, b_; /* the rgb value when the gradient stops. */

    gdouble witness; /* Where does we must show the line to say the values ^^ ? */

    cairo_pattern_t* pat; /* This pattern will be used to make some gradient. */

    /* We create a linear gradient, which has the width of the widget. */
    pat = cairo_pattern_create_linear (0, 0, ECSC_WIDTH, 0);


    witness = ella_color_selection_channel_get_value(color)*(ECSC_WIDTH - 5);

    r   = color->red;
    r_  = color->red;
    g   = color->green;
    g_  = color->green;
    b   = color->blue;
    b_  = color->blue;

    switch(color->channel)
    {
        case ELLA_COLOR_CHANNEL_RED:
        {
            r = 0;
            r_ = 1;
            break;
        }
        case ELLA_COLOR_CHANNEL_GREEN:
        {
            g = 0;
            g_ = 1;
            break;
        }
        case ELLA_COLOR_CHANNEL_BLUE:
        {
            b = 0;
            b_ = 1;
            break;
        }
    }

    /* Create the main gradient */
    cairo_pattern_add_color_stop_rgba(pat, 0, r, g, b, 1);
    cairo_pattern_add_color_stop_rgba(pat, 1, r_, g_, b_, 1);

    /* Fill our rectangle, with our gradient. */

    cairo_rectangle(cr, 0, 0, ECSC_WIDTH, gtk_widget_get_allocated_height(widget));
    cairo_set_source_rgb(cr, 0, 0, 0);
    cairo_fill(cr);

    cairo_rectangle(cr, 1, 1, ECSC_WIDTH - 2, gtk_widget_get_allocated_height(widget) - 2);
    cairo_set_source (cr, pat);
    cairo_fill(cr);

    cairo_pattern_destroy (pat);

    /* Get the default color of the user. */
    theme_color = ella_config_get_gtk_color(widget);

    cairo_set_source_rgb(cr, (gdouble)theme_color->red/0xFFFF,
                             (gdouble)theme_color->green/0xFFFF,
                             (gdouble)theme_color->blue/0xFFFF);
    cairo_rectangle(cr, 1 + witness, 3, 2, gtk_widget_get_allocated_height(widget)-6);
    cairo_fill(cr);
    return TRUE;
}

static void
ella_color_selection_channel_init (EllaColorSelectionChannel *color)
{
    gtk_widget_add_events(GTK_WIDGET (color),
                          GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
                          GDK_POINTER_MOTION_MASK);
    gtk_widget_set_size_request(GTK_WIDGET(color), ELLA_COLOR_SELECTION_CHANNEL_WIDTH, ELLA_COLOR_SELECTION_CHANNEL_HEIGHT);
}

/**
 * \brief Create a new EllaColorSelectionChannel.
 **/
GtkWidget* ella_color_selection_channel_new(void)
{
    return g_object_new(ELLA_TYPE_COLOR_SELECTION_CHANNEL, NULL);
}

/**
 * \brief Set the main channel of the widget.
 * \param ecsc A EllaColorSelectionChannel
 * \param type The main channel: red, green or blue.
 **/
void ella_color_selection_channel_set_channel(EllaColorSelectionChannel* ecsc, EllaColorChannelType type)
{
    ecsc->channel = type;
}

/**
 * \brief Set the red value of the EllaColorSelectionChannel
 * \param ecs A EllaColorSelectionChannel
 * \param value The new value of the red.
 **/
void ella_color_selection_channel_set_value_r(EllaColorSelectionChannel* color, gdouble value)
{
    color->red = value;
    gtk_widget_queue_draw(GTK_WIDGET(color));
}

/**
 * \brief Set the green value of the EllaColorSelectionChannel
 * \param ecs A EllaColorSelectionChannel
 * \param value The new value of the green.
 **/
void ella_color_selection_channel_set_value_g(EllaColorSelectionChannel* color, gdouble value)
{
    color->green = value;
    gtk_widget_queue_draw(GTK_WIDGET(color));
}

/**
 * \brief Set the blue value of the EllaColorSelectionChannel
 * \param ecs A EllaColorSelectionChannel
 * \param value The new value of the blue.
 **/
void ella_color_selection_channel_set_value_b(EllaColorSelectionChannel* color, gdouble value)
{
    color->blue = value;
    gtk_widget_queue_draw(GTK_WIDGET(color));
}

/**
 * \brief Get the blue value of the EllaColorSelectionChannel, if it main
 * channel is blue or green, this value may not change without an other
 * interaction (e.g. with ella_color_selection_set_value_r).
 * \param ecs A EllaColorSelectionChannel
 * \return The red value of the color selected.
 **/
gdouble ella_color_selection_channel_get_value_r(EllaColorSelectionChannel* color)
{
    return color->red;
}

/**
 * \brief Get the green value of the EllaColorSelectionChannel, if it main
 * channel is red or blue, this value may not change without an other
 * interaction (e.g. with ella_color_selection_set_value_g).
 * \param ecs A EllaColorSelectionChannel
 * \return The green value of the color selected.
 **/
gdouble ella_color_selection_channel_get_value_g(EllaColorSelectionChannel* color)
{
    return color->green;
}

/**
 * \brief Get the blue value of the EllaColorSelectionChannel, if it main
 * channel is red or green, this value may not change without an other
 * interaction (e.g. with ella_color_selection_set_value_b).
 * \param ecs A EllaColorSelectionChannel
 * \return The blue value of the color selected.
 **/
gdouble ella_color_selection_channel_get_value_b(EllaColorSelectionChannel* ecs)
{
    return ecs->blue;
}

/**
 * \param ecs A EllaColorSelectionChannel
 * \return The value of the EllaColorSelectionChannel, it depends on it channel
 *         e.g. if it is a red channel, it returns the red value.
 **/
gdouble ella_color_selection_channel_get_value(EllaColorSelectionChannel* ecs)
{
    switch(ecs->channel)
    {
        case ELLA_COLOR_CHANNEL_RED:
            return ecs->red;
        case ELLA_COLOR_CHANNEL_GREEN:
            return ecs->green;
        case ELLA_COLOR_CHANNEL_BLUE:
            return ecs->blue;
        default:
            g_error("Couldn't get the channel of this widget");
    }
}
