/* Copyright (C) 2010 Lucas Baudin
 *
 * This file is part of Ella Project.
 *
 * Ella Project is free software : you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Ella Project is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY ; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ella Project. If not, see <http ://www.gnu.org/licenses/>.
 */

#include "widgets/properties_panel.h"
#include "ella-color_selection.h"

/* Callbacks for properties panel widget, and for the canvas */
static void on_canvas_selection_changed(EllaCanvas* canvas, EllaUIPanel* panel);
static void ella_properties_panel_connect_signals(EllaUIPanel* panel);

EllaUIPanel* ella_properties_panel_new(GtkBuilder* builder)
{
    EllaUIPanel* panel = g_malloc(sizeof(EllaUIPanel));

    /* Load the widgets from the builder  to hide/show them later */
    panel->expander_text     = GTK_WIDGET(gtk_builder_get_object(builder,
                                                                 "expander_edit_text"));
    panel->expander_stroke   = GTK_WIDGET(gtk_builder_get_object(builder,
                                                                 "expander_stroke"));
    panel->hseparator_fill   = GTK_WIDGET(gtk_builder_get_object(builder,
                                                                 "hseparator_fill"));
    panel->hseparator_stroke = GTK_WIDGET(gtk_builder_get_object(builder,
                                                                 "hseparator_stroke"));
    panel->hseparator_text   = GTK_WIDGET(gtk_builder_get_object(builder,
                                                                 "hseparator_text"));
    panel->line_width        = GTK_WIDGET(gtk_builder_get_object(builder,
                                                                 "w_spinbuttonPropertiesLineWidth"));
    panel->txt_buffer        = GTK_TEXT_BUFFER(gtk_builder_get_object(builder,
                                                                      "textbuffer1"));

    panel->show_stroke = TRUE;
    panel->show_text = TRUE;

    /* Create our nice ECS */
    panel->ecs = ella_color_selection_new();
    panel->ecs_stroke = ella_color_selection_new();

    /* And add them to the panel */
    gtk_box_pack_start(GTK_BOX(gtk_builder_get_object(builder, "vbox_fill")), panel->ecs, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(gtk_builder_get_object(builder, "vbox_stroke_style")), panel->ecs_stroke, FALSE, FALSE, 0);

    ella_properties_panel_connect_signals(panel);

    return panel;
}

void ella_properties_panel_update(EllaUIPanel* panel)
{
    if(panel->expander_text)
    {
        /* gtk_widget_set_visible has been introduced in gtk 2.18, so it doesn't
         * work with old version of gtk.*/
        if(panel->show_text)
        {
            gtk_widget_show(panel->expander_text);
            gtk_widget_show(panel->hseparator_text);
        }
        else
        {
            gtk_widget_hide(panel->expander_text);
            gtk_widget_hide(panel->hseparator_text);
        }
    }
    else
    {
        g_warning("Couldn't load the expander text widget.");
    }

    if(panel->expander_stroke)
    {
        /* gtk_widget_set_visible has been introduced in gtk 2.18, so it doesn't
         * work with old version of gtk.*/
        if(panel->show_stroke)
        {
            gtk_widget_show(panel->expander_stroke);
            gtk_widget_show(panel->hseparator_stroke);
        }
        else
        {
            gtk_widget_hide(panel->expander_stroke);
            gtk_widget_hide(panel->hseparator_stroke);
        }
    }
    else
    {
        g_warning("Couldn't load the expander stroke widget.");
    }

}

void ella_properties_panel_set(EllaUIPanel* panel,
                               EllaUIPanelField widget,
                               gboolean visible)
{
    /* What field do we want to hide or show? */
    switch(widget)
    {
        case ELLA_UI_PANEL_TEXT:
        {
            panel->show_text = visible;
            break;
        }
        case ELLA_UI_PANEL_STROKE:
        {
            panel->show_stroke = visible;
            break;
        }
        default:
        {
            g_warning("Couldn't understand what widgets you want to hide or show.");
        }
    }
}

static void ella_properties_panel_connect_signals(EllaUIPanel* panel)
{
  g_signal_connect(canvas, "selection_changed",
                   G_CALLBACK (on_canvas_selection_changed),
                   panel);
}

static void on_canvas_selection_changed(EllaCanvas* canvas, EllaUIPanel* panel)
{
    EllaColor selection_color;
    EllaColor selection_stroke_color;
    double line_width;

    /* Update the text (of the first item selected), if no item is selected,
     * this function returns "" */
    gtk_text_buffer_set_text(panel->txt_buffer,
                             ella_canvas_get_selection_text(canvas),
                             -1 /* our gchar is null terminated */);

    ella_properties_panel_set(panel, ELLA_UI_PANEL_TEXT, ella_canvas_is_text_item_selected(canvas));

    /* Get the color of the first item selected, if no item is selected, a
     * default value is returned */
    ella_canvas_get_selection_color(canvas, &selection_color);

    /* Get the color of the first item selected, if no item is selected, a
     * default value is returned */
    ella_canvas_get_selection_stroke_color(canvas, &selection_stroke_color);

    line_width = ella_canvas_get_selection_line_width(canvas);

    /* Set the value in our ECS */
    ella_color_selection_set_value_rgba(ELLA_COLOR_SELECTION(panel->ecs),
                                        selection_color.red,
                                        selection_color.green,
                                        selection_color.blue,
                                        selection_color.alpha);

    /* Set the value in our ECS */
    ella_color_selection_set_value_rgba(ELLA_COLOR_SELECTION(panel->ecs_stroke),
                                        selection_stroke_color.red,
                                        selection_stroke_color.green,
                                        selection_stroke_color.blue,
                                        selection_stroke_color.alpha);

    /* Now line width */
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(panel->line_width), line_width);
    
    ella_properties_panel_update(panel);
}

void on_edit_text_changed(GtkWidget* txt_buffer, gpointer item)
{
    GtkTextIter start;
    GtkTextIter end;

    gchar *text;

    /* Obtain iters for the start and end of points of the buffer */
    gtk_text_buffer_get_start_iter(GTK_TEXT_BUFFER(txt_buffer), &start);
    gtk_text_buffer_get_end_iter(GTK_TEXT_BUFFER(txt_buffer), &end);

    /* Get the entire buffer text. */
    text = gtk_text_buffer_get_text(GTK_TEXT_BUFFER(txt_buffer), &start, &end, FALSE);

    ella_canvas_set_selection_text(canvas, text);

    g_free (text);
}

void on_fonts_changed(GtkComboBox* combo, gpointer item)
{
    /* Just check that it is athe good object */
    g_return_if_fail(GTK_IS_COMBO_BOX(combo));

    /* Get the valid iter and get the good value from it */
    GtkTreeModel* model = gtk_combo_box_get_model(combo);
    GtkTreeIter iter;
    gtk_combo_box_get_active_iter(combo, &iter);
    GValue value = {0};
    gtk_tree_model_get_value(model, &iter, 0, &value);

    /* Put the new font in our canvas */
    g_debug("Font changed: %s\n", g_value_get_string(&value));
    ella_canvas_set_selection_font_name(canvas, g_value_get_string(&value));
}
